#
#
#           The Nim Compiler
#        (c) Copyright 2017 Andreas Rumpf
#
#    See the file "copying.txt", included in this
#    distribution, for details about the copyright.
#

# This include implements the new semantic prepass so we can ban
# forward decls from Nim for good.

proc determineProcType(c: PContext; n: PNode; s: PSym) =
  pushOwner(s)
  openScope(c)
  var gp: PNode
  if n.sons[genericParamsPos].kind != nkEmpty:
    n.sons[genericParamsPos] = semGenericParamList(c, n.sons[genericParamsPos])
    gp = n.sons[genericParamsPos]
  else:
    gp = newNodeI(nkGenericParams, n.info)
  # process parameters:
  if n.sons[paramsPos].kind != nkEmpty:
    semParamList(c, n.sons[paramsPos], gp, s)
    if sonsLen(gp) > 0:
      if n.sons[genericParamsPos].kind == nkEmpty:
        # we have a list of implicit type parameters:
        n.sons[genericParamsPos] = gp
        # check for semantics again:
        # semParamList(c, n.sons[ParamsPos], nil, s)
  else:
    s.typ = newProcType(c, n.info)
  if tfTriggersCompileTime in s.typ.flags: incl(s.flags, sfCompileTime)
  if n.sons[patternPos].kind != nkEmpty:
    n.sons[patternPos] = semPattern(c, n.sons[patternPos])
  if s.kind == skIterator:
    s.typ.flags.incl(tfIterator)
  if s.ast.isNil: s.ast = n
  closeScope(c)
  popOwner()

proc registerRoutineDecl(c: PContext; n: PNode; kind: TSymKind) =
  case c.phase
  of stepRegisterSymbol:
    if n[namePos].kind notin {nkEmpty, nkSym}:
      let s = semIdentDef(c, n.sons[0], kind)
      n.sons[namePos] = newSymNode(s)

      addInterfaceOverloadableSymAt(c, c.currentScope, s)
      s.flags.incl sfForward
  of stepDetermineType:
    if n[namePos].kind == nkSym:
      let s = n[namePos].sym
      s.owner = getCurrOwner()
      if s.typ == nil:
        determineProcType(c, n, s)
  of stepCompileBody, stepDone:
    discard "traditional pass implemented in the legacy"

proc importModulePass1(up: SharedContext; importingModule: PSym;
                       fileIdx: int32; isCyclic: bool): PSym {.
                       nimcall, procvar.} =
  if isCyclic:
    result = up.importer(up, importingModule, fileIdx, isCyclic)

proc semConstPrepassName(c: PContext; n: PNode) =
  for i in countup(0, sonsLen(n) - 1):
    let a = n.sons[i]
    if a.kind == nkCommentStmt: continue
    if (a.kind != nkConstDef): illFormedAst(a)
    checkSonsLen(a, 3)
    let v = semIdentDef(c, a.sons[0], skConst)
    if sfGenSym notin v.flags: addInterfaceDecl(c, v)
    a.sons[0] = newSymNode(v)

proc semConstPrepassType(c: PContext; n: PNode) =
  for i in countup(0, sonsLen(n) - 1):
    let a = n.sons[i]
    if gCmd == cmdIdeTools: suggestStmt(c, a)
    if a.kind == nkCommentStmt: continue
    if (a.kind != nkConstDef): illFormedAst(a)

    checkSonsLen(a, 3)
    if a[0].kind != nkSym: continue
    var typ: PType = nil
    if a.sons[1].kind != nkEmpty: typ = semTypeNode(c, a.sons[1], nil)

    var def = semConstExpr(c, a.sons[2])
    if def == nil:
      localError(a.sons[2].info, errConstExprExpected)
      continue
    # check type compatibility between def.typ and typ:
    if typ != nil:
      def = fitRemoveHiddenConv(c, typ, def)
    else:
      typ = def.typ
    if typ == nil:
      localError(a.sons[2].info, errConstExprExpected)
      continue
    if typeAllowed(typ, skConst) != nil and def.kind != nkNilLit:
      localError(a.info, errXisNoType, typeToString(typ))
      continue
    var v = a[0].sym
    v.typ = typ
    v.ast = def
    a.sons[2] = def

proc prePass*(c: PContext; n: PNode) =
  # first pass: add all global symbols to the symbol table.
  #   special case: type T = someTempl(X) is not
  # second pass: resolve proc signatures
  # third pass: resolve proc bodies (the traditional pass that
  # requires var/let to come before the proc).
  case n.kind
  of nkConstSection:
    if c.phase == stepRegisterSymbol:
      semConstPrepassName(c, n)
    elif c.phase == stepDetermineType:
      semConstPrepassType(c, n)
  of nkTypeSection:
    case c.phase
    of stepRegisterSymbol:
      incl n.flags, nfSem
      typeSectionLeftSidePass(c, n)
    of stepDetermineType: typeSectionRightSidePass(c, n)
    of stepCompileBody: typeSectionFinalPass(c, n)
    of stepDone: discard
  of nkUsingStmt:
    if c.phase == stepRegisterSymbol:
      discard semUsing(c, n)
  of nkIteratorDef: registerRoutineDecl(c, n, skIterator)
  of nkProcDef: registerRoutineDecl(c, n, skProc)
  of nkMethodDef: registerRoutineDecl(c, n, skMethod)
  of nkConverterDef: registerRoutineDecl(c, n, skConverter)
  of nkMacroDef: registerRoutineDecl(c, n, skMacro)
  of nkTemplateDef: registerRoutineDecl(c, n, skTemplate)
  of nkImportStmt:
    if c.phase == stepRegisterSymbol:
      for i in countup(0, sonsLen(n) - 1):
        var m = triggerModuleImport(c, n.sons[i], importModulePass1)
        if m != nil:
          #importForwarded(c, m.ast, emptySet)
          n.sons[i] = newSymNode(m)
    elif c.phase == stepDetermineType:
      var emptySet: IntSet
      for it in n:
        if it.kind == nkSym:
          let m = it.sym
          # ``addDecl`` needs to be done before ``importAllSymbols``!
          addDecl(c, m)             # add symbol to symbol table of module
          importAllSymbolsExcept(c, m, emptySet)
  of nkImportExceptStmt:
    if c.phase == stepRegisterSymbol:
      checkMinSonsLen(n, 2)
      var m = triggerModuleImport(c, n.sons[0], importModulePass1)
      if m != nil:
        n.sons[0] = newSymNode(m)
    elif c.phase == stepDetermineType and n.len > 0 and n.sons[0].kind == nkSym:
      let m = n.sons[0].sym
      addDecl(c, m)               # add symbol to symbol table of module
      var exceptSet = initIntSet()
      for i in countup(1, sonsLen(n) - 1):
        let ident = lookups.considerQuotedIdent(n.sons[i])
        exceptSet.incl(ident.id)
      importAllSymbolsExcept(c, m, exceptSet)
  of nkFromStmt:
    if c.phase == stepRegisterSymbol:
      checkMinSonsLen(n, 2)
      var m = triggerModuleImport(c, n.sons[0], importModulePass1)
      if m != nil:
        n.sons[0] = newSymNode(m)
    elif c.phase == stepDetermineType and n.len > 0 and n.sons[0].kind == nkSym:
      let m = n.sons[0].sym
      addDecl(c, m)               # add symbol to symbol table of module
      for i in countup(1, sonsLen(n) - 1):
        if n.sons[i].kind != nkNilLit:
          importSymbol(c, n.sons[i], m)
  of nkIncludeStmt:
    discard evalInclude(c, n)
  of nkExportStmt, nkExportExceptStmt:
    discard semExport(c, n)
  of nkStmtList:
    for i in 0 .. < n.len:
      prePass(c, n.sons[i])
  else: discard
